import pickle


def pickle_object(obj, path):
	"""
	Stores an object as a file.
	Parameters
	----------
	obj: The object instance
	path: Full path as a str where the file will be stored.

	Returns
	-------

	"""
	with open(path, "wb") as f:
		pickle.dump(obj, f)


def read_pickle(path):
	"""
	Reads a file containing a pickled object and returns it
	Parameters
	----------
	path: Full path as a str where the file is stored.

	Returns an object.
	-------

	"""
	with open(path, "rb") as f:
		return pickle.load(f)


def open_file(path, mode):
	with open(path, mode) as f:
		return f.read()



import abc
from cobamp.nullspace.subset_reduction import subset_reduction
from cobamp.utilities.property_management import PropertyDictionary
from cobamp.core.models import ConstraintBasedModel
from numpy import array, nonzero, ndarray
from itertools import chain, product
from copy import deepcopy


class ModelTransformer(object):
	__metaclass__ = abc.ABCMeta

	def transform(self, args, properties):

		# args must be:
		# - a dict with 'S', 'lb', 'ub' keys
		# - a ConstraintBasedModel
		if isinstance(args, dict):
			assert len(set(args.keys()) & {'S', 'lb', 'ub'}) == len(set(args.keys())), 'args must contain at least S' + \
																					   ', lb, and ub key-value pairs'

			S, lb, ub = [args[k] for k in ['S', 'lb', 'ub']]
			return self.transform_array(S, lb, ub, properties)

		elif isinstance(args, ConstraintBasedModel):
			S = args.get_stoichiometric_matrix()
			lb, ub = args.get_bounds_as_list()
			new_properties = deepcopy(properties)

			for k in ['block','keep']:
				if new_properties[k] != None:
					new_properties[k] = [args.decode_index(r,'reaction') for r in properties[k]]

			Sn, lbn, ubn, mapping, metabs = self.transform_array(S, lb, ub, new_properties)

			reaction_names_new = [new_properties['reaction_id_sep'].join([args.reaction_names[i] for i in mapping.from_new(i)]) for i in
								  range(len(lbn))]
			modeln = ConstraintBasedModel(
				S=Sn,
				thermodynamic_constraints=[list(k) for k in list(zip(lbn, ubn))],
				reaction_names=reaction_names_new,
				metabolite_names= [args.metabolite_names[k] for k in metabs]
			)

			return modeln, mapping, metabs

	@abc.abstractmethod
	def transform_array(self, S, lb, ub, properties):
		## TODO: implement

		# must return:
		# - new S matrix
		# - new lower/upper bounds
		# - mapping between rows/cols from both matrices
		# mapping = ReactionIndexMapping({}, {})
		# metabs = []
		# return S, lb, ub, mapping, metabs
		return


class ReactionIndexMapping(object):
	def __init__(self, otn, nto):
		self.otn = otn
		self.nto = nto

	def from_original(self, idx):
		return self.otn[idx]

	def from_new(self, idx):
		return self.nto[idx]

	def multiply(self, new_ids):
		return list(product(*[self.from_new(k) for k in set(new_ids)]))


class SubsetReducerProperties(PropertyDictionary):
	def __init__(self, keep=None, block=None, absolute_bounds=False, reaction_id_sep='_+_'):

		def is_list(x):
			return isinstance(x, (tuple, list, ndarray))

		new_optional = {
			'keep': lambda x: is_list(x) or type(x) == None,
			'block': lambda x: is_list(x) or type(x) == None,
			'absolute_bounds': bool,
			'reaction_id_sep': str
		}

		super().__init__(optional_properties=new_optional)
		for name, value in zip(['keep', 'block', 'absolute_bounds', 'reaction_id_sep'],
							   [keep, block, absolute_bounds, reaction_id_sep]):
			self.add_if_not_none(name, value)


class SubsetReducer(ModelTransformer):
	TO_KEEP_SINGLE = 'SUBSET_REDUCER-TO_KEEP_SINGLE'
	TO_BLOCK = 'SUBSET_REDUCER-TO_BLOCK'
	ABSOLUTE_BOUNDS = 'SUBSET_REDUCER-ABSOLUTE_BOUNDS'

	def reduce(self, S, lb, ub, keep=(), block=(), absolute_bounds=False):
		lb, ub = list(map(array, [lb, ub]))
		to_keep, to_block = [], []
		irrev = (lb >= 0) | ((ub <= 0) & (lb <= 0))

		if block:
			to_block = array(block)

		if keep:
			to_keep = array(keep)

		rd, sub, irrev_reduced, rdind, irrv_subsets, kept_reactions, K, _ = subset_reduction(
			S, irrev, to_keep_single=to_keep, to_remove=to_block)

		mapping = self.get_transform_maps(sub)

		nlb = [0 if irrev_reduced[k] else None for k in range(rd.shape[1])]
		nub = [None] * rd.shape[1]

		if absolute_bounds:
			nlb = [0 if irrev_reduced[k] else -float('inf') for k in range(rd.shape[1])]
			nub = [float('inf')] * rd.shape[1]
			alb, aub = list(zip(*[[fx([x[k] for k in mapping.from_new(i)]) for x, fx in zip([lb, ub], [max, min])]
								  for i in range(rd.shape[1])]))

			for func, pair in zip([max, min], [[nlb, alb], [nub, aub]]):
				new, absolute = pair
				for i, v in enumerate(absolute):
					new[i] = func(new[i], absolute[i])

		return rd, nlb, nub, mapping, rdind

	def transform_array(self, S, lb, ub, properties):
		k, b, a = (properties[k] for k in ['keep', 'block', 'absolute_bounds'])

		Sn, lbn, ubn, mapping, metabs = self.reduce(S, lb, ub, k, b, a)

		return Sn, lbn, ubn, mapping, metabs

	def get_transform_maps(self, sub):
		new_to_orig = {i: list(nonzero(sub[i, :])[0]) for i in range(sub.shape[0])}
		orig_to_new = dict(chain(*[[(i, k) for i in v] for k, v in new_to_orig.items()]))

		return ReactionIndexMapping(orig_to_new, new_to_orig)



from collections import OrderedDict
from itertools import chain
import abc
import numpy as np
import warnings

import optlang
from optlang.symbolics import Zero

CUSTOM_DEFAULT_SOLVER = None
SOLVER_ORDER = ['CPLEX', 'GUROBI', 'GLPK', 'MOSEK', 'SCIPY']


def get_default_solver():
	if CUSTOM_DEFAULT_SOLVER:
		return CUSTOM_DEFAULT_SOLVER
	else:
		for solver in SOLVER_ORDER:
			if optlang.list_available_solvers()[solver]:
				return solver


def get_solver_interfaces():
	s = {}
	for solver, status in optlang.list_available_solvers().items():
		if status:
			s[solver] = eval('optlang.' + solver.lower() + '_interface')
	return s


SOLVER_INTERFACES = get_solver_interfaces()
DEFAULT_SOLVER = get_default_solver()

VAR_CONTINUOUS, VAR_INTEGER, VAR_BINARY = ('continuous', 'integer', 'binary')
VAR_TYPES = (VAR_CONTINUOUS, VAR_INTEGER, VAR_BINARY)
VARIABLE, CONSTRAINT = 'var', 'const'
SENSE_MINIMIZE, SENSE_MAXIMIZE = ('min', 'max')

def fwd_irrev(lb, ub):
	return ((lb >= 0) and (ub >= 0)).astype(int)

def bak_irrev(lb, ub):
	return ((lb < 0) and (ub <= 0)).astype(int)


def fix_backwards_irreversible_reactions(S, lb, ub):
	lb, ub = np.array(lb), np.array(ub)
	fwd_irrev_index, bak_irrev_index = [[i for i in range(S.shape[1]) if fx(lb[i], ub[i])] for fx in [fwd_irrev, bak_irrev]]
	# irrev = np.union1d(fwd_irrev_index, bak_irrev_index)

	if len(bak_irrev_index) > 0:
		S[:, bak_irrev_index] = -S[:, bak_irrev_index]
		ub_temp = ub[bak_irrev_index]
		ub[bak_irrev_index] = -lb[bak_irrev_index]
		lb[bak_irrev_index] = -ub_temp

	return S, lb, ub, fwd_irrev_index, bak_irrev_index

def make_irreversible_model(S, lb, ub):
	# lb, ub = np.array(lb), np.array(ub)
	# # fwd_irrev = lambda lb, ub: (lb >= 0) and (ub >= 0)
	# # bak_irrev = lambda lb, ub: (lb < 0) and (ub <= 0)
	#
	# fwd_irrev_index, bak_irrev_index = [[i for i in range(S.shape[1]) if fx(lb[i], ub[i])] for fx in [fwd_irrev, bak_irrev]]
	S, lb, ub, fwd_irrev_index, bak_irrev_index = fix_backwards_irreversible_reactions(S, lb, ub)

	irrev = np.union1d(fwd_irrev_index, bak_irrev_index)

	# irrev = np.array([i for i in range(S.shape[1]) if not (lb[i] < 0 and ub[i] > 0)])
	rev = np.array([i for i in range(S.shape[1]) if i not in irrev])

	# if len(bak_irrev_index) > 0:
	# 	S[:, bak_irrev_index] = -S[:, bak_irrev_index]
	# 	ub_temp = ub[bak_irrev_index]
	# 	ub[bak_irrev_index] = -lb[bak_irrev_index]
	# 	lb[bak_irrev_index] = -ub_temp

	Sr = S[:, rev]
	offset = S.shape[1]
	rx_mapping = {k: v if k in irrev else [v] for k, v in dict(zip(range(offset), range(offset))).items()}
	for i, rx in enumerate(rev):
		rx_mapping[rx].append(offset + i)
	rx_mapping = OrderedDict([(k, tuple(v)) if isinstance(v, list) else (k, v) for k, v in rx_mapping.items()])

	S_new = np.hstack([S, -Sr])
	nlb, nub = np.zeros(S_new.shape[1]), np.zeros(S_new.shape[1])
	for orig_rx, new_rx in rx_mapping.items():
		if isinstance(new_rx, tuple):
			nub[new_rx[0]] = abs(lb[orig_rx])
			nub[new_rx[1]] = ub[orig_rx]
		else:
			nlb[new_rx], nub[new_rx] = lb[orig_rx], ub[orig_rx]

	return S_new, nlb, nub, rx_mapping


class LinearSystem():
	"""
	An abstract class defining the template for subclasses implementing linear systems that can be used with optimizers
	such as CPLEX and passed onto other algorithms supplied with the package.

	Must instantiate the following variables:

		S: System of linear equations represented as a n-by-m ndarray, preferrably with dtype as float or int

		__model: Linear model as an instance of the solver.
	"""
	__metaclass__ = abc.ABCMeta
	model = None

	def select_solver(self, solver=None):
		if not solver:
			solver = get_default_solver()
			self.interface = SOLVER_INTERFACES[solver]
		else:
			if solver in SOLVER_INTERFACES:
				self.interface = SOLVER_INTERFACES[solver]

			else:
				raise Exception(solver, 'Solver not found. Choose from ', str(list(SOLVER_INTERFACES.keys())))
		self.solver = solver

	def get_configuration(self):
		return self.model.configuration

	def set_default_configuration(self):
		cfg = self.model.configuration
		try:
			cfg.tolerances.feasibility = 1e-9
			cfg.tolerances.optimality = 1e-9
			cfg.tolerances.integrality = 1e-9
			cfg.lpmethod = 'auto'
			cfg.presolve = True

		except:
			print('Could not set parameters with this solver')

	@abc.abstractmethod
	def build_problem(self):
		"""
		Builds a CPLEX model with the constraints specified in the constructor arguments.
		This method must be implemented by any <LinearSystem>.
		Refer to the :ref:`constructor <self.__init__>`
		-------

		"""
		pass

	def get_model(self):
		"""

		Returns the model instance. Must call <self.build_problem()> to return a CPLEX model.

		-------

		"""
		return self.model

	def get_stoich_matrix_shape(self):
		"""

		Returns a tuple with the shape (rows, columns) of the supplied stoichiometric matrix.

		-------

		"""
		return self.S.shape

	# def empty_constraint(self, b_lb, b_ub, dummy_var=Variable(name='dummy', lb=0, ub=0)):
	def empty_constraint(self, b_lb, b_ub, **kwargs):
		return self.interface.Constraint(Zero, lb=b_lb, ub=b_ub, **kwargs)

	def dummy_variable(self):
		return self.interface.Variable(name='dummy', lb=0, ub=0)

	def populate_model_from_matrix(self, S, var_types, lb, ub, b_lb, b_ub, var_names, only_nonzero=False,
								   indicator_rows=None):
		'''

		Args:
			model: Optlang model
			S: Two-dimensional np.ndarray instance
			var_types: list or tuple with length equal to the amount of columns of S containing variable types (e.g.
			VAR_CONTINUOUS)
			lb: list or tuple with length equal to the amount of columns of S containing the lower bounds for all variables
			ub: list or tuple with length equal to the amount of columns of S containing the upper bounds for all variables
			b_lb: list or tuple with length equal to the amount of rows of S containing the lower bound for the right hand
			side of all constraints
			b_ub: list or tuple with length equal to the amount of rows of S containing the upper bound for the right hand
			side of all constraints
			var_names: string identifiers for the variables

		'''

		self.add_variables_to_model(var_names, lb, ub, var_types)
		self.add_rows_to_model(S, b_lb, b_ub, only_nonzero, indicator_rows)

	def populate_constraints_from_matrix(self, S, constraints, vars, only_nonzero=False):
		'''

		Args:
			S: Two-dimensional np.ndarray instance
			b_lb: list or tuple with length equal to the amount of rows of S containing the lower bound for the right hand
			side of all constraints
			b_ub: list or tuple with length equal to the amount of rows of S containing the upper bound for the right hand
			side of all constraints

			vars: list of variable instances

		'''
		if not only_nonzero:
			coef_list = [{vars[j]: S[i, j] for j in range(S.shape[1])} for i in range(S.shape[0])]
		else:
			coef_list = [{vars[j]: S[i, j] for j in np.nonzero(S[i, :])[0]} for i in range(S.shape[0])]

		for coefs, constraint in zip(coef_list, constraints):
			if constraint.indicator_variable is None:
				constraint.set_linear_coefficients(coefs)

		self.model.update()

	def add_rows_to_model(self, S_new, b_lb, b_ub, only_nonzero=False, indicator_rows=None, vars=None, names=None):
		if not vars:
			vars = self.model.variables
		# dummy = self.dummy_variable()
		if names != None:
			constraints = [self.empty_constraint(b_lb[i], b_ub[i], name=names[i]) for i in range(S_new.shape[0])]
		else:
			constraints = [self.empty_constraint(b_lb[i], b_ub[i]) for i in range(S_new.shape[0])]
		if indicator_rows:
			for row, var_idx, complement in indicator_rows:
				constraints[row] = self.interface.Constraint(
					sum(S_new[row, i] * vars[i] for i in S_new[row, :].nonzero()[0]), lb=b_lb[row], ub=b_ub[row],
					indicator_variable=vars[var_idx], active_when=complement)
		self.model.add(constraints, sloppy=True)

		self.model.update()

		self.populate_constraints_from_matrix(S_new, constraints, vars, only_nonzero)
		# self.model.update()
		# self.model.remove(dummy)
		self.model.update()
		return constraints

	def remove_from_model(self, index, what):
		container = self.model.variables if what == VARIABLE else self.model.constraints if what == CONSTRAINT else None
		if type(index) not in (list, tuple):
			index = [index]
		for i in index:
			self.model.remove(container[i])

		if what == VARIABLE:
			self.S = np.delete(self.S, index, 1)
		elif what == CONSTRAINT:
			self.S = np.delete(self.S, index, 0)

		self.model.update()

	def add_columns_to_model(self, S_new, var_names, lb, ub, var_types):
		vars = self.add_variables_to_model(var_names, lb, ub, var_types)

		self.populate_constraints_from_matrix(S_new, self.model.constraints, vars)

	def add_variables_to_model(self, var_names, lb, ub, var_types):

		if isinstance(var_types, str):
			var_types = [var_types] * len(lb)

		vars = [self.interface.Variable(name=name, lb=lbv, ub=ubv, type=t) for name, lbv, ubv, t in
				zip(var_names, lb, ub, var_types)]
		self.model.add(vars)
		# debug lines - TODO: remove this
		# for var in vars:
		# 	print(var)
		self.model.update()

		return vars

	def set_objective(self, coefficients, minimize, vars=None):
		if not vars:
			vars = self.model.variables
		lcoefs = {k: v for k, v in zip(vars, coefficients) if v != 0}
		dummy = self.dummy_variable()
		new_obj = self.interface.Objective(dummy)
		self.model.objective = new_obj
		new_obj.set_linear_coefficients(lcoefs)
		self.model.remove(dummy)
		self.model.objective.direction = SENSE_MINIMIZE if minimize else SENSE_MAXIMIZE
		self.model.update()

	def set_variable_bounds(self, vars, lb, ub):
		for var, ulb, uub in zip(vars, lb, ub):
			var.set_bounds(ulb, uub)
		self.model.update()

	def set_constraint_bounds(self, constraints, lb, ub):
		for c, ulb, uub in zip(constraints, lb, ub):
			b = ulb, uub
			lb_is_greater = b[0] > c.ub if ((c.ub is not None) and (b[0] is not None)) else False
			ub_is_lower = b[1] < c.lb if ((c.lb is not None) and (b[1] is not None)) else False
			if lb_is_greater:
				c.ub = b[1]
				c.lb = b[0]
			elif ub_is_lower:
				c.lb = b[0]
				c.ub = b[1]
			else:
				c.lb, c.ub = b
		self.model.update()

	def set_variable_types(self, vars, types):
		if isinstance(types, str):
			for var in vars:
				var.type = types
		else:
			for var, typ in zip(vars, types):
				if typ in VAR_TYPES:
					var.type = typ
				else:
					warnings.warn('Invalid variable type: ' + typ)
		self.model.update()

	def write_to_lp(self, filename):
		with open(filename, 'w') as f:
			f.write(self.model.to_lp())


class KShortestCompatibleLinearSystem(LinearSystem):
	"""
	Abstract class representing a linear system that can be passed as an argument for the KShortestAlgorithm class.
	Subclasses must instantiate the following variables:

		__dvar_mapping: A dictionary mapping reaction indexes with variable names

		__dvars: A list of variable names (str) or Tuple[str] if two linear system variables represent a single flux.
		Should be kept as type `list` to maintain order.

		__c: str representing the variable to be used as constant for indicator constraints
	"""
	dvar_mapping = None
	dvars = None
	c = None

	def get_dvar_mapping(self):
		"""

		Returns a dictionary mapping flux indices with variable(s) on the optimization problem
		-------

		"""
		return self.dvar_mapping

	def get_dvars(self):
		"""

		Returns a list of variables (str or Tuple[str]) with similar order to that of the fluxes passed to the system.
		-------

		"""
		return self.dvars

	def add_c_variable(self):
		self.c = self.add_variables_to_model(['C'], [1], [None], VAR_CONTINUOUS)[0]

	def get_c_variable(self):
		return self.model.variables['C']


class GenericLinearSystem(LinearSystem):
	"""
	Class representing a generic system of linear (in)equations
	Used as arguments for various algorithms implemented in the package.
	"""

	def __init__(self, S, var_types, lb, ub, b_lb, b_ub, var_names, solver=None):
		"""
		Constructor for GenericLinearSystem

		Parameters

			model: Optlang model
			S: Two-dimensional np.ndarray instance
			var_types: list or tuple with length equal to the amount of columns of S containing variable types (e.g.
			VAR_CONTINUOUS)
			lb: list or tuple with length equal to the amount of columns of S containing the lower bounds for all variables
			ub: list or tuple with length equal to the amount of columns of S containing the upper bounds for all variables
			b_lb: list or tuple with length equal to the amount of rows of S containing the lower bound for the right hand
			side of all constraints
			b_ub: list or tuple with length equal to the amount of rows of S containing the upper bound for the right hand
			side of all constraints
			var_names: string identifiers for the variables

		"""
		self.select_solver(solver)
		self.model = self.interface.Model()
		self.set_default_configuration()
		self.S, self.lb, self.ub, self.b_lb, self.b_ub, self.var_types = S, lb, ub, b_lb, b_ub, var_types

		self.names = var_names if var_names is not None else ['v' + str(i) for i in range(S.shape[1])]

	def build_problem(self):
		self.populate_model_from_matrix(self.S, self.var_types, self.lb, self.ub, self.b_lb, self.b_ub, self.names,
										True)


class SteadyStateLinearSystem(GenericLinearSystem):
	"""
	Class representing a steady-state biological system of metabolites and reactions without dynamic parameters
	Used as arguments for various algorithms implemented in the package.
	"""

	def __init__(self, S, lb, ub, var_names, solver=None):
		"""
		Constructor for SimpleLinearSystem

		Parameters

		----------

			S: Stoichiometric matrix represented as a n-by-m ndarray, preferrably with dtype as float or int

			lb: ndarray or list containing the lower bounds for all n fluxes

			ub: ndarray or list containing the lower bounds for all n fluxes

			var_names: - optional - ndarray or list containing the names for each flux
		"""
		m, n = S.shape
		self.lb, self.ub = lb, ub
		super().__init__(S, VAR_CONTINUOUS, lb, ub, [0] * m, [0] * m, var_names, solver=solver)


class IrreversibleLinearSystem(KShortestCompatibleLinearSystem, GenericLinearSystem):
	"""
	Class representing a steady-state biological system of metabolites and reactions without dynamic parameters.
	All irreversible reactions are split into their forward and backward components so every lower bound is 0.
	Used as arguments for various algorithms implemented in the package.
	"""

	def __init__(self, S, lb, ub, non_consumed=(), consumed=(), produced=(), solver=None, force_bounds={}):
		"""

		Parameters
		----------

			S: Stoichiometric matrix represented as a n-by-m ndarray, preferrably with dtype as float or int

			irrev: An Iterable[int] or ndarray containing the indices of irreversible reactions

			non_consumed: An Iterable[int] or ndarray containing the indices of external metabolites not consumed in the
			model.

			consumed: An Iterable[int] or ndarray containing the indices of external metabolites guaranteed to be produced.

			produced: An Iterable[int] or ndarray containing the indices of external metabolites guaranteed to be consumed.
		"""

		self.select_solver(solver)

		# lb = [0 if i in irrev else -1 for i in range(S.shape[1])]
		# ub = [1] * S.shape[1]
		# if -1 in lb:
		S, lb, ub, fwd_irrev, bak_irrev = fix_backwards_irreversible_reactions(S, lb, ub)

		Si, lbi, ubi, rev_mapping = make_irreversible_model(S, lb, ub)
		fwd_names = ['V' + str(i) if not isinstance(v, list) else 'V' + str(i) + 'F' for i, v in rev_mapping.items()]
		bwd_names = ['V' + str(i) + 'R' for i, v in rev_mapping.items() if isinstance(v, (list, tuple))]

		# else:
		# 	Si, lbi, ubi = S, lb, ub
		# 	fwd_names = ['V' + str(i) for i in range(Si.shape[1])]
		# 	bwd_names = []
		# 	rev_mapping = {i:i for i in range(Si.shape[1])}

		lbi = [0] * len(lbi) + [1]
		ubi = [None] * len(ubi) + [None]

		for k,tup in force_bounds.items():
			if isinstance(rev_mapping[k], (list,tuple)):
				inds = rev_mapping[k]
				trev = (abs(ub[k]) if ub[k] < 0 else 0, abs(lb[k]) if lb[k] < 0 else 0)
				tfwd = (abs(lb[k]) if lb[k] > 0 else 0, abs(ub[k]) if ub[k] > 0 else 0)
				assert sum([pair[0] < pair[1] for pair in (trev,tfwd)]) == 2, 'force_bounds contains invalid values'
				for i, ntup in zip(inds,(tfwd,trev)):
					lbi[i],ubi[i] = ntup
			else:
				assert tup[0] < tup[1], 'force_bounds contains invalid values'
				lbi[k],ubi[k] = tup

		self.rev_mapping = rev_mapping
		self.__ivars = None
		self.__ss_override = [(nc, 'G', 0) for nc in non_consumed] + [(p, 'G', 1) for p in produced] + [(c, 'L', -1) for
																										c in consumed]
		Si = np.hstack([Si, np.zeros((Si.shape[0], 1))])
		b_lb, b_ub = [0] * Si.shape[0], [0] * Si.shape[0]

		## TODO: Maybe allow other values to be provided for constraint relaxation/tightening

		for id, _, v in self.__ss_override:
			b_lb[id], b_ub[id] = (v, None) if v >= 0 else (None, v)

		super().__init__(Si, VAR_CONTINUOUS, lbi, ubi, b_lb, b_ub, fwd_names + bwd_names + ['C'], solver=solver)

		self.dvars = list(range(S.shape[1] + len(bwd_names)))
		self.dvar_mapping = dict(rev_mapping)


class IrreversibleLinearPatternSystem(IrreversibleLinearSystem):
	## TODO: Code + docstrings. Do not use this yet!
	def __init__(self, S, lb, ub, subset, **kwargs):
		super().__init__(S, lb, ub, **kwargs)
		self.subset = subset
		fwds, revs = [], []
		dvm_new = {}
		for r in self.subset:
			if not isinstance(self.rev_mapping[r], int):
				fid, rid = self.dvar_mapping[r]
				dvm_new[r] = (len(fwds),len(subset)+len(revs))
				fwds.append(fid)
				revs.append(rid)
			else:
				dvm_new[r] = len(fwds)
				fwds.append(self.dvar_mapping[r])

		self.dvar_mapping = dvm_new
		self.dvars = fwds + revs

	def build_problem(self):
		super().build_problem()


class DualLinearSystem(KShortestCompatibleLinearSystem, GenericLinearSystem):
	"""
	Class representing a dual system based on a steady-state metabolic network whose elementary flux modes are minimal
	cut sets for use with the KShortest algorithms. Based on previous work by Ballerstein et al. and Von Kamp et al.
	References:
	[1] von Kamp, A., & Klamt, S. (2014). Enumeration of smallest intervention strategies in genome-scale metabolic
	networks. PLoS computational biology, 10(1), e1003378.
	[2] Ballerstein, K., von Kamp, A., Klamt, S., & Haus, U. U. (2011). Minimal cut sets in a metabolic network are
	elementary modes in a dual network. Bioinformatics, 28(3), 381-387.
	"""

	def __init__(self, S, lb, ub, T, b, solver=None):
		"""

		Parameters

		----------

			S: Stoichiometric matrix represented as a n-by-m ndarray, preferrably with dtype as float or int
			irrev: An Iterable[int] or ndarray containing the indices of irreversible reactions

			T: Target matrix as an ndarray. Should have c-by-n dimensions (c - #constraints; n - #fluxes)

			b: Inhomogeneous bound values as a list or 1D ndarray of c size n.
		"""
		self.select_solver(solver)

		S, lb, ub, fwd_irrev, bak_irrev = fix_backwards_irreversible_reactions(S, lb, ub)

		irrev = np.union1d(fwd_irrev, bak_irrev).astype(int)

		self.__ivars = None
		self.S, self.irrev, self.T, self.b = S, irrev, T, b
		self.__c = "C"
		super().__init__(*self.generate_dual_problem(S, irrev, T, b), solver=solver)

		offset = S.shape[0]
		self.dvars = list(range(offset, offset + (S.shape[1] * 2)))

		self.dvar_mapping = {i: (i, S.shape[1] + i) for i in range(S.shape[1])}

	def generate_dual_problem(self, S, irrev, T, b):
		m, n = S.shape
		veclens = [("u", m), ("vp", n), ("vn", n), ("w", self.T.shape[0])]
		I = np.identity(n)
		Sxi, Sxr = S[:, irrev].T, np.delete(S, irrev, axis=1).T
		Ii, Ir = I[irrev, :], np.delete(I, irrev, axis=0)
		Ti, Tr = T[:, irrev].T, np.delete(T, irrev, axis=1).T

		u, vp, vn, w = [[(pref + str(i), 0 if pref != "u" else None, None) for i in range(n)] for
						pref, n in
						veclens]
		Sdi = np.hstack([Sxi, Ii, -Ii, Ti, np.zeros([Sxi.shape[0], 1])])
		Sdr = np.hstack([Sxr, Ir, -Ir, Tr, np.zeros([Sxr.shape[0], 1])])
		Sd = np.vstack([Sdi, Sdr, np.zeros([1, Sdi.shape[1]])])
		names, v_lb, v_ub = list(zip(*list(chain(u, vp, vn, w))))

		names = list(names) + ['C']
		v_lb = list(v_lb) + [1]
		v_ub = list(v_ub) + [None]

		w_idx = [m + n + n + i for i in range(len(b))]
		Sd[-1, w_idx] = b
		Sd[-1, -1] = 1

		b_ub = np.hstack([np.array([None] * Sdi.shape[0]), np.array([0] * Sdr.shape[0]), np.array([0])])
		b_lb = np.array([0] * (Sd.shape[0]), dtype=object)
		b_ub[-1] = 0
		b_lb[-1] = None

		return Sd, VAR_CONTINUOUS, v_lb, v_ub, b_lb, b_ub, names



from ..utilities.tree import Tree

def compress_linear_paths(tree):
	"""
	Collapses sequences of nodes contained in a Tree with only one children as a single node containing all values of
	those nodes.
	Parameters
	----------
	tree: A Tree instance.
	-------

	"""
	if tree.is_leaf():
		pass
	else:
		if len(tree.children) == 1:
			if type(tree.value) != list:
				tree.value = [tree.value]
			tree.value.append(tree.children[0].value)
			tree.children = tree.children[0].children
			compress_linear_paths(tree)
		for child in tree.children:
			compress_linear_paths(child)


def ignore_compressed_nodes_by_size(tree, size):
	"""
	Modifies the values of a tree's children that have been previously compressed with the <compress_linear_paths>
	function if they contain more than a certain number of elements. The node's value is changed to "REMAINING".
	Parameters
	----------
	tree: A Tree instance
	size: An integer with the size threshold
	-------

	"""
	for child in tree.children:
		if isinstance(child.value, list) and len(child.value) > size:
			child.value = 'REMAINING'
			child.children = []
		else:
			ignore_compressed_nodes_by_size(child, size)


def probabilistic_tree_prune(tree, target_level, current_level=0, cut_leaves=False, name_separator=' and '):
	"""
	Cuts a tree's nodes under a certain height (`target_level`) and converts ensuing nodes into a single one whose value
	represents the relative frequency of an element in the nodes below. Requires values on the extra_info field.
	Parameters
	----------
	tree: A Tree instance
	target_level: An int representing the level at which the tree will be cut
	current_level: The current level of the tree (int). Default is 0 for root nodes.
	cut_leaves: A boolean indicating whether the node at the target level is excluded or displays probabilities.
	name_separator: Separator to use when representing multiple elements
	-------

	"""
	if current_level < target_level:
		if target_level == current_level and cut_leaves:
			tree.children = []
		else:
			for child in tree.children:
				probabilistic_tree_prune(child, target_level, current_level + 1, cut_leaves, name_separator)

	else:
		probabilistic_tree_compression(tree, name_separator=name_separator)
		tree.value = "REMAINING" if cut_leaves else [str(k) + "=" + str(tree.value[k]) for k in
													 sorted(tree.value, key=tree.value.get)]
	return target_level == current_level


def probabilistic_tree_compression(tree, data=None, total_count=None, name_separator=' and '):
	"""
	Compresses a node and subsequent children by removing them and modifying the value to a dictionary with the relative
	frequency of each element in the subsequent nodes. Requires values on the extra_info field.

	Parameters
	----------
	tree: A Tree instance
	data: Local count if not available in extra_info
	total_count: Total amount of sets if not available in extra_info
	name_separator: Separator to use when representing multiple elements
	-------

	"""
	if data is None and total_count is None:
		total_count = int(tree.extra_info)
		data = {name_separator.join(tree.value) if isinstance(tree.value, list) else tree.value: 1}
		for child in tree.children:
			probabilistic_tree_compression(child, data, total_count, name_separator)
		tree.value = data
		tree.children = []
	else:
		local_proportion = int(tree.extra_info) / total_count
		key = name_separator.join(tree.value) if isinstance(tree.value, list) else tree.value
		if key not in data.keys():
			data[key] = local_proportion
		else:
			data[key] += local_proportion
		for child in tree.children:
			probabilistic_tree_compression(child, data, total_count, name_separator)


def pretty_print_tree(tree, write_path=None):
	"""
	Parameters
	----------
	tree: A Tree instance
	write_path: Path to store a text file. Use None if the string is not to be stored in a file.

	Returns a text representation of a Tree instance along with its children.
	-------

	"""
	buffer = []

	def __pretty_print_tree(tree, buffer, overhead, final_node=True):
		current_line = overhead + "|-- " + repr(tree)
		buffer.append(current_line)
		for child in tree.children:
			__pretty_print_tree(child, buffer, overhead + "|\t", False)
		if final_node:
			return buffer

	__pretty_print_tree(tree, buffer, '', True)

	res = '\n'.join(buffer)

	if write_path is not None:
		with open(write_path, 'w') as f:
			f.write(res)

	return res


def apply_fx_to_all_node_values(tree, fx):
	"""
	Applies a function to all nodes below the tree, modifying their value to its result.
	Parameters
	----------
	tree: A Tree instance
	fx: A function to apply
	-------

	"""
	tree.value = fx(tree.value)
	for child in tree.children:
		apply_fx_to_all_node_values(child, fx)


def find_all_tree_nodes(tree):
	"""
	Parameters
	----------
	tree: A Tree instance.

	Returns a list of all nodes below a node
	-------

	"""

	def __find_all_tree_nodes(tree, it):
		it.append(tree)
		for child in tree.children:
			__find_all_tree_nodes(child, it)

	it = []
	__find_all_tree_nodes(tree, it)
	return it


def merge_duplicate_nodes(tree):
	"""
	Merges all nodes with similar values, replacing every instance reference of all nodes with the same object if its
	value is identical

	Parameters
	----------
	tree: A Tree instance
	-------

	"""
	all_nodes = find_all_tree_nodes(tree)
	conv_key = lambda x: str(sorted(x) if isinstance(x, list) else x)
	unique_keys = [conv_key(k.value) for k in all_nodes]
	unique_node_map = {k: Tree(k) for k in unique_keys}

	def __merge_duplicate_nodes(tree):
		new_children = []
		for child in tree.children:
			grandchildren = child.children
			new_child = unique_node_map[conv_key(child.value)]
			new_child.children = grandchildren
			new_children.append(new_child)
		tree.children = new_children
		for child in tree.children:
			__merge_duplicate_nodes(child)

	__merge_duplicate_nodes(tree)


def populate_nx_graph(tree, G, previous=None, name_separator='\n', unique_nodes=True, node_dict=None):
	if node_dict is None:
		node_dict = {}
	if unique_nodes:
		node_value_key = name_separator.join(tree.value) if type(tree.value) == list else str(tree.value)
		node_value = node_value_key
		if node_value_key not in node_dict.keys():
			node_dict[node_value_key] = 1
			node_value = node_value + "_" + '0'
		else:
			node_value = node_value + "_" + str(node_dict[node_value])
			node_dict[node_value_key] += 1
	else:
		node_value = tree.value
		node_value_key = tree.value
	if previous != None:
		previous_node, previous_key = previous
		G.add_edge(previous_node, node_value)
	if not tree.is_leaf():
		for child in tree.children:
			populate_nx_graph(child, G, previous=(node_value, node_value_key), name_separator=name_separator,
							  unique_nodes=unique_nodes, node_dict=node_dict)



# from .algorithms import kshortest_efms, kshortest_efm_properties
# from .linear_systems import linear_systems, optimization
# from .mcs_enumeration import intervention_problem
# from .nullspace import subset_reduction, nullspace
# from .utilities import file_utils, property_management, set_utils, tree_analysis, external_wrappers

from . import algorithms, analysis, core, nullspace, utilities, wrappers



from numpy import ndarray, array, delete, zeros, vstack, hstack, nonzero, append, int_, int8, int16, \
	int32, int64
from .linear_systems import SteadyStateLinearSystem, VAR_CONTINUOUS, make_irreversible_model
from .optimization import LinearSystemOptimizer, CORSOSolution, GIMMESolution
from collections import OrderedDict
import warnings
from copy import deepcopy

INT_TYPES = (int, int_, int8, int16, int32, int64)
LARGE_NUMBER = 10e6 - 1
SMALL_NUMBER = 1e-6
BACKPREFIX = 'flux_backwards'


def make_irreversible_model_raven(S, lb, ub, inverse_reverse_reactions=False):
	irrev = array([i for i in range(S.shape[1]) if not (lb[i] < 0)])
	rev = array([i for i in range(S.shape[1]) if i not in irrev])
	Sr = S[:, rev]
	offset = S.shape[1]
	rx_mapping = {k: v if k in irrev else [v] for k, v in dict(zip(range(offset), range(offset))).items()}
	for i, rx in enumerate(rev):
		rx_mapping[rx].append(offset + i)
	rx_mapping = OrderedDict([(k, tuple(v)) if isinstance(v, list) else (k, v) for k, v in rx_mapping.items()])

	S_new = hstack([S, -Sr])
	nlb, nub = zeros(S_new.shape[1]), zeros(S_new.shape[1])
	for orig_rx, new_rx in rx_mapping.items():
		if isinstance(new_rx, tuple):
			nlb[new_rx[0]] = lb[orig_rx] if lb[orig_rx] >= 0 else 0  # not necessary since they're already 0
			nub[new_rx[0]] = ub[orig_rx] if ub[orig_rx] >= 0 else 0  # not necessary since they're already 0
			nlb[new_rx[1]] = ub[orig_rx] * -1 if (ub[orig_rx] * -1) >= 0 else 0
			nub[new_rx[1]] = lb[orig_rx] * -1 if (lb[orig_rx] * -1) >= 0 else 0

		else:
			nlb[new_rx] = lb[orig_rx]
			nub[new_rx] = ub[orig_rx]

	return S_new, nlb, nub, rx_mapping




# def make_irreversible_model(S, lb, ub):
# 	irrev = array([i for i in range(S.shape[1]) if not (lb[i] < 0 and ub[i] > 0)])
# 	rev = array([i for i in range(S.shape[1]) if i not in irrev])
# 	Sr = S[:, rev]
# 	offset = S.shape[1]
# 	rx_mapping = {k: v if k in irrev else [v] for k, v in dict(zip(range(offset), range(offset))).items()}
# 	for i, rx in enumerate(rev):
# 		rx_mapping[rx].append(offset + i)
# 	rx_mapping = OrderedDict([(k, tuple(v)) if isinstance(v, list) else (k, v) for k, v in rx_mapping.items()])
#
# 	S_new = hstack([S, -Sr])
# 	nlb, nub = zeros(S_new.shape[1]), zeros(S_new.shape[1])
# 	for orig_rx, new_rx in rx_mapping.items():
# 		if isinstance(new_rx, tuple):
# 			nub[new_rx[0]] = abs(lb[orig_rx])
# 			nub[new_rx[1]] = ub[orig_rx]
# 		else:
# 			nlb[new_rx], nub[new_rx] = lb[orig_rx], ub[orig_rx]
#
# 	return S_new, nlb, nub, rx_mapping


# def test_irreversible_conversions():
# 	S = zeros([3,3])
# 	lb, ub = ([0,1000],[-1000, 1000],[-1000, 0])

class ConstraintBasedModel(object):
	def __init__(self, S, thermodynamic_constraints, reaction_names=None, metabolite_names=None, optimizer=True,
				 solver=None):

		def __validate_args():

			# stoichiometric matrix
			if isinstance(S, ndarray):
				m, n = S.shape
			elif isinstance(S, tuple) or isinstance(S, list):
				m, n = len(S), len(S[0])
			else:
				raise TypeError('S matrix is not an ndarray or list')

			thermodynamic_types = set(map(type, thermodynamic_constraints))
			allowed_types = {list, tuple, bool}

			assert len(thermodynamic_types & allowed_types) == len(thermodynamic_types), 'Invalid bound type found.'

			reactions_ok = reaction_names is None or len(reaction_names) == n and isinstance(reaction_names, list)
			metabolites_ok = metabolite_names is None or len(metabolite_names) == m and isinstance(metabolite_names,
																								   list)

			assert reactions_ok, 'Reaction name dimensions do not match the supplied stoichiometrix matrix'
			assert metabolites_ok, 'Metabolite name dimensions do not match the supplied stoichiometrix matrix'
			return m, n

		self.solver = solver
		m, n = __validate_args()
		# self.__S = self.__parse_stoichiometric_matrix(S)
		self.__S = array(S)

		self.bounds = self.__interpret_bounds(thermodynamic_constraints)
		self.original_bounds = deepcopy(self.bounds)
		self.reaction_names, self.metabolite_names = deepcopy(reaction_names), deepcopy(metabolite_names)

		self.__update_decoder_map()

		self.c = None
		self.model = None

		if optimizer:
			self.initialize_optimizer()

	def __getstate__(self):
		return self.__dict__

	def __update_decoder_map(self):
		self.reaction_decoder_map = self.metabolite_decoder_map = None

		if self.reaction_names:
			self.reaction_decoder_map = dict(zip(self.reaction_names, list(range(self.__S.shape[1]))))
		if self.metabolite_names:
			self.metabolite_decoder_map = dict(zip(self.metabolite_names, list(range(self.__S.shape[0]))))

		self.map_labels = {'reaction': self.reaction_decoder_map,
						   'metabolite': self.metabolite_decoder_map}

	def __interpret_bounds(self, thermodynamic_constraints):
		def interpret_bound(bound_obj):
			i, bound = bound_obj
			#			value = [None, None]
			if isinstance(bound, bool):
				if bound:
					value = 0, LARGE_NUMBER
				else:
					value = -LARGE_NUMBER, LARGE_NUMBER
			else:
				value = bound
				for i, v in enumerate(value):
					if i == 0 and v is None:
						value[i] = -LARGE_NUMBER
					elif i == 1 and v is None:
						value[i] = LARGE_NUMBER
			if value[0] > value[1]:
				warnings.warn('Lower bound is >= than upper bound for reaction ' + str(i) + '. Reversing...')
				value = value[::-1]
			return value

		return list(map(interpret_bound, enumerate(thermodynamic_constraints)))

	def flux_limits(self, reaction):
		self.set_objective({reaction:1}, True)
		min_flux = self.optimize().objective_value()
		self.set_objective({reaction:1}, False)
		max_flux = self.optimize().objective_value()

		return min_flux, max_flux

	def initialize_optimizer(self):
		lb, ub = self.get_bounds_as_list()
		self.model = SteadyStateLinearSystem(self.__S, lb, ub, var_names=self.reaction_names, solver=self.solver)
		self.model.build_problem()
		self.optimizer = LinearSystemOptimizer(self.model, build=False)

	def decode_index(self, index, labels):
		if type(index) in INT_TYPES:
			return index

		elif labels is not None:
			return self.map_labels[labels][index]
		else:
			raise IndexError('Could not find specified index "' + str(index) + '".')

	def get_bounds_as_list(self):
		return list(zip(*self.bounds))

	def is_reversible_reaction(self, index):
		lb, ub = self.bounds[self.decode_index(index, 'reaction')]
		return (lb < 0 and ub > 0)

	def get_stoichiometric_matrix(self, rows=None, columns=None):
		row_index = [self.decode_index(i, 'metabolite') for i in rows] if rows else None
		col_index = [self.decode_index(i, 'reaction') for i in columns] if columns else None

		if rows and columns:
			return self.__S[row_index, col_index]
		elif rows:
			return self.__S[row_index, :]
		elif columns:
			return self.__S[:, col_index]
		else:
			return self.__S

	def set_stoichiometric_matrix(self, values, rows=None, columns=None):
		row_index = [self.decode_index(i, 'metabolite') for i in rows] if rows else None
		col_index = [self.decode_index(i, 'reaction') for i in columns] if columns else None

		if rows and columns:
			self.__S[row_index, col_index] = values
		elif rows:
			self.__S[row_index, :] = values
		elif columns:
			self.__S[:, col_index] = values
		else:
			self.__S = values

		row_index = row_index if row_index is not None else range(self.__S.shape[0])
		col_index = col_index if col_index is not None else range(self.__S.shape[1])

		if self.model:
			constraints = [self.model.model.constraints[i] for i in
						   row_index] if row_index else self.model.model.constraints
			vars = [self.model.model.variables[i] for i in col_index] if row_index else self.model.model.variables

			self.model.populate_constraints_from_matrix(values, constraints=constraints, vars=vars)

	def add_metabolite(self, arg, name=None):
		assert not name in self.metabolite_names, 'Duplicate metabolite name found!'
		if isinstance(arg, dict):
			row = zeros(1, self.__S.shape[1])
			for k, v in arg.items():
				row[self.decode_index(k, 'reaction')] = v
		elif isinstance(arg, ndarray):
			if len(arg) == len(self.reaction_names):
				row = arg
			else:
				raise Exception('Numpy argument dimensions should be (', len(self.metabolite_names), ',). Got',
								arg.shape, 'instead')
		else:
			raise ValueError('Invalid argument type: ', type(arg), '. Please supply an ndarray or dict instead.')
		self.__S = vstack([self.__S, row])
		self.metabolite_names.append(name)

		self.__update_decoder_map()

		if self.model:
			self.model.add_rows_to_model(row.reshape([1, self.__S.shape[1]]), b_lb=array([0]), b_ub=array([0]))

	def add_reaction(self, arg, bounds, name=None):
		assert not name in self.reaction_names, 'Duplicate reaction name found!'
		if isinstance(arg, dict):
			col = zeros([self.__S.shape[0], 1])
			for k, v in arg.items():
				col[self.decode_index(k, 'metabolite'),0] = v
		elif isinstance(arg, ndarray):
			if len(arg) == len(self.metabolite_names):
				col = arg.reshape(self.__S.shape[0], 1)
			else:
				raise Exception('Numpy argument dimensions should be (', len(self.reaction_names), ',). Got', arg.shape,
								'instead')
		else:
			raise ValueError('Invalid argument type: ', type(arg), '. Please supply an ndarray or dict instead.')
		self.__S = hstack([self.__S, col])

		self.reaction_names.append(name)
		self.bounds.append(bounds)

		self.__update_decoder_map()

		if self.model:
			self.model.add_columns_to_model(col, [name], [bounds[0]], [bounds[1]], VAR_CONTINUOUS)

	def remove_reaction(self, index):
		j = self.decode_index(index, 'reaction')
		if not isinstance(index, int):
			self.reaction_names.pop(j)
		self.bounds.pop(j)
		self.__S = delete(self.__S, [j], axis=1)
		self.__update_decoder_map()

		if self.model:
			self.model.remove_from_model(j, 'var')

	def remove_metabolite(self, index):
		i = self.decode_index(index, 'metabolite')
		if not isinstance(index, int):
			self.metabolite_names.pop(i)

		self.__S = delete(self.__S, [i], axis=0)
		self.__update_decoder_map()

		if self.model:
			self.model.remove_from_model(i, 'const')

	def make_irreversible(self):
		lb, ub = self.get_bounds_as_list()
		Sn, nlb, nub, rx_mapping = make_irreversible_model(self.__S, lb, ub)
		rev = array([int(v[0]) for k, v in rx_mapping.items() if isinstance(v, (list, tuple))])
		revnames = ['_'.join([name, BACKPREFIX]) for name in (array(self.reaction_names)[rev]).tolist()]
		rnames = self.reaction_names + revnames
		model = ConstraintBasedModel(Sn, list(zip(nlb, nub)), rnames, self.metabolite_names,
									 optimizer=self.model is not None, solver=self.solver)
		return model, rx_mapping

	def get_reaction_bounds(self, index):
		return self.bounds[self.decode_index(index, 'reaction')]

	def set_reaction_bounds(self, index, **kwargs):
		true_idx = self.decode_index(index, 'reaction')
		lb, ub = self.get_reaction_bounds(true_idx)
		if 'lb' in kwargs:
			lb = kwargs['lb']
		if 'ub' in kwargs:
			ub = kwargs['ub']
		bound = (lb, ub)
		self.bounds[true_idx] = bound

		if 'temporary' in kwargs and kwargs['temporary'] == False:
			self.original_bounds[self.decode_index(index, 'reaction')] = self.get_reaction_bounds(index)

		if self.model:
			var = self.model.model.variables[true_idx]
			self.model.set_variable_bounds([var], [lb], [ub])

	def set_objective(self, coef_dict, minimize=False):
		if self.model:
			if isinstance(coef_dict, dict):
				f = zeros(self.__S.shape[1], )
				self.c = f
				for k, v in coef_dict.items():
					self.c[self.decode_index(k, 'reaction')] = v
				self.model.set_objective(self.c, minimize)
			elif isinstance(coef_dict, ndarray):
				self.model.set_objective(coef_dict, minimize)
			else:
				raise TypeError('`coef_dict` must either be a dict or an ndarray')
		else:
			raise Exception('Cannot set an objective on a model without the optimizer flag as True.')

	def optimize(self, coef_dict=None, minimize=False):
		cur_obj = self.model.model.objective
		if coef_dict != None:
			self.set_objective(coef_dict, minimize)
		sol = self.optimizer.optimize()

		if coef_dict != None:
			self.model.model.objective = cur_obj

		return sol


	def revert_to_original_bounds(self):
		for rx, bounds in zip(self.reaction_names, self.original_bounds):
			clb, cub = self.get_reaction_bounds(rx)
			lb, ub = bounds

			if clb != lb or cub != ub:
				self.set_reaction_bounds(rx, lb=lb, ub=ub)


class CORSOModel(ConstraintBasedModel):
	def __init__(self, cbmodel, corso_element_names=('R_PSEUDO_CORSO', 'M_PSEUDO_CORSO'), solver=None):
		if not cbmodel.model:
			cbmodel.initialize_optimizer()

		self.cbmodel = cbmodel

		irrev_model, self.mapping = cbmodel.make_irreversible()

		S = irrev_model.get_stoichiometric_matrix()
		bounds = irrev_model.bounds

		self.cost_index_mapping = zeros(S.shape[1], dtype=int_)

		self.corso_rx, self.corso_mt = corso_element_names
		super().__init__(S, bounds, irrev_model.reaction_names, irrev_model.metabolite_names, solver=solver)
		self.add_metabolite(zeros(len(self.reaction_names)), self.corso_mt)

		self.add_reaction(zeros(len(self.metabolite_names)), (0, 0), self.corso_rx)

		self.original_bounds = deepcopy(self.bounds)

		for orx, nrx in self.mapping.items():
			if isinstance(nrx, int):
				self.cost_index_mapping[nrx] = orx
			else:
				for nrx_split in nrx:
					self.cost_index_mapping[nrx_split] = orx

	def solve_original_model(self, of_dict, minimize=False):
		self.cbmodel.set_objective(of_dict, minimize)
		sol = self.cbmodel.optimize()
		return sol

	def set_costs(self, cost):
		true_cost = cost[self.cost_index_mapping]
		true_cost = append(true_cost, array([-1]))
		self.set_stoichiometric_matrix(true_cost.reshape(1, len(true_cost)), rows=[self.corso_mt])

	# def set_reaction_bounds(self, index, **kwargs):
	# 	super().set_reaction_bounds(index, **kwargs)
	# 	self.original_bounds[self.decode_index(index, 'reaction')] = self.get_reaction_bounds(index)

	def set_corso_objective(self):
		self.set_objective({self.corso_rx: 1}, True)

	def optimize_corso(self, cost, of_dict, minimize=False, constraint=1, constraintby='val', eps=1e-6, flux1=None):
		if flux1 is None:
			flux1 = self.solve_original_model(of_dict, minimize)

		if abs(flux1.objective_value()) < eps:
			return flux1, flux1
		f1_x = flux1.x()
		if constraintby == 'perc':
			# f1_f = flux1.x()[self.cbmodel.c != 0]
			f1_f = {idx: f1_x[idx] * (constraint / 100) for idx in of_dict.keys()}
		elif constraintby == 'val':
			if (flux1.objective_value() < constraint and not minimize) or (
					flux1.objective_value() > constraint and minimize):
				raise Exception('Objective flux is not sufficient for the the set constraint value.')
			else:
				f1_f = {idx: constraint for idx in of_dict.keys()}
		else:
			raise Exception('Invalid constraint options.')

		self.set_reaction_bounds(self.corso_rx, lb=0, ub=1e20)
		# corso_of_dict = deepcopy(of_dict)
		# corso_of_dict[self.corso_rx] = 1

		self.set_costs(cost)
		for rx in f1_f.keys():
			true_idx = self.decode_index(rx, 'reaction')
			involved = self.mapping[true_idx]
			fluxval = f1_f[rx]  # if isinstance(f1_f, ndarray) else f1_f

			if isinstance(involved, (int, int_)):
				self.set_reaction_bounds(involved, lb=fluxval, ub=fluxval, temporary=True)
			else:
				self.set_reaction_bounds(involved[0], lb=fluxval, ub=fluxval, temporary=True)
				self.set_reaction_bounds(involved[1], lb=0, ub=0, temporary=True)

		self.set_objective({self.corso_rx: 1}, True)

		sol = self.optimize()
		self.revert_to_original_bounds()

		return flux1, CORSOSolution(flux1, sol, sum([of_dict[k] * f1_f[k] for k in of_dict.keys()]),
									self.cost_index_mapping, self.cbmodel.reaction_names, eps=eps)


class GIMMEModel(ConstraintBasedModel):
	def __init__(self, cbmodel, solver=None):
		self.cbmodel = cbmodel
		if not self.cbmodel.model:
			self.cbmodel.initialize_optimizer()

		irrev_model, self.mapping = cbmodel.make_irreversible()

		S = irrev_model.get_stoichiometric_matrix()
		bounds = irrev_model.bounds
		super().__init__(S, bounds, irrev_model.reaction_names, irrev_model.metabolite_names, solver=solver,
						 optimizer=True)

	def __adjust_objective_to_irreversible(self, objective_dict):
		obj_dict = {}
		for k, v in objective_dict.items():
			irrev_map = self.mapping[self.cbmodel.decode_index(k, 'reaction')]
			if isinstance(irrev_map, (list, tuple)):
				for i in irrev_map:
					obj_dict[i] = v
			else:
				obj_dict[irrev_map] = v
		return obj_dict

	def __adjust_expression_vector_to_irreversible(self, exp_vector):
		exp_vector_n = zeros(len(self.reaction_names), )
		for rxn, val in enumerate(exp_vector):
			rxmap = self.mapping[rxn]
			if isinstance(rxmap, tuple):
				exp_vector_n[rxmap[0]] = exp_vector_n[rxmap[1]] = val
			else:
				exp_vector_n[rxmap] = val
		return exp_vector_n

	def optimize_gimme(self, exp_vector, objectives, obj_frac, flux_thres):
		N = len(self.cbmodel.reaction_names)
		objectives_irr = [self.__adjust_objective_to_irreversible(obj) for obj in objectives]
		exp_vector_irr = self.__adjust_expression_vector_to_irreversible(exp_vector)

		def find_objective_value(obj):
			self.cbmodel.set_objective(obj, False)
			return self.cbmodel.optimize().objective_value()

		objective_values = list(map(find_objective_value, objectives_irr))

		gimme_model_objective = array(
			[flux_thres - exp_vector_irr[i] if -1 < exp_vector_irr[i] < flux_thres else 0 for i in range(N)])

		objective_lbs = zeros(len(self.reaction_names))
		for ov, obj in zip(objective_values, objectives_irr):
			for rx, v in obj.items():
				objective_lbs[rx] = v * ov * obj_frac

		objective_ids = nonzero(objective_lbs)[0]
		for id, lb in zip(objective_ids, objective_lbs):
			self.set_reaction_bounds(id, lb=lb, temporary=True)

		self.set_objective(gimme_model_objective, True)
		sol = self.optimize()
		self.revert_to_original_bounds()
		return GIMMESolution(sol, exp_vector, self.cbmodel.reaction_names, self.mapping)



import time


def timeit(method):
	'''
	Timer decorator for methods. Courtesy of Fahim Sakri from PythonHive
	Args:
		method:

	Returns:

	'''

	def timed(*args, **kw):
		ts = time.time()
		result = method(*args, **kw)
		te = time.time()

		if 'log_time' in kw:
			name = kw.get('log_name', method.__name__.upper())
			kw['log_time'][name] = int((te - ts) * 1000)
		else:
			print('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
		return result

	return timed



'''
	Implementation of the fast flux variability analysis method by Gudmundsson and Thiele (Computationally efficient
	flux variability analysis - BMC Bioinformatics 2010 11:489).

	Some parts have been heavily inspired by the talented people supporting cobrapy.

'''

from cobamp.core.optimization import LinearSystemOptimizer
from numpy import zeros
from pathos.pools import _ProcessPool
from pathos.multiprocessing import cpu_count

def _fva_initializer(linear_system, sense, gamma):
	global _linear_system
	global _opt
	global _sense
	global _lenrx
	global _gamma
	_linear_system = linear_system
	_lenrx = _linear_system.get_stoich_matrix_shape()[1]
	_opt = LinearSystemOptimizer(_linear_system, build=False)
	_sense = sense
	_gamma = gamma


def _fva_iteration(i):
	global _linear_system
	global _opt
	global _sense
	global _gamma
	global _lenrx
	#print('Iterating for ',i)
	w = zeros( _lenrx).ravel()
	w[i] = 1
	_linear_system.set_objective(w, _sense)
	sol = _opt.optimize()
	#_linear_system.add_rows_to_model(w, [sol.objective_value()* _gamma], [None], only_nonzero=True)
	# print(w, _sense, i, sol.x()[i], sol.objective_value())
	return i, sol.objective_value()



class FluxVariabilityAnalysis(object):
	def __init__(self, linear_system, workers=None):
		self.ls = linear_system
		self.n_jobs = min(cpu_count() if workers ==	None else workers, linear_system.get_stoich_matrix_shape()[1])

	def run(self, initial_objective, minimize_initial, gamma=1-1e-6):

		M,N = self.ls.get_stoich_matrix_shape()
		result = {i:[0,0] for i in range(N)}
		opt = LinearSystemOptimizer(self.ls, build=False)
		c = zeros(N)
		c[initial_objective] = 1
		self.ls.set_objective(c, minimize_initial)

		v0 = opt.optimize()
		z0 = v0.objective_value()

		self.ls.add_rows_to_model(c.reshape([1, N]), [z0 * gamma], [None], only_nonzero=True,
										 names=['FASTFVAINITIALCONSTRAINT'])

		for sense in [True, False]:
			rx_per_job = N // self.n_jobs
			self.pool = _ProcessPool(
				processes=self.n_jobs,
				initializer=_fva_initializer,
				initargs=(self.ls, sense, gamma)
			)
			for i, value in self.pool.imap_unordered(_fva_iteration, range(N), chunksize=rx_per_job):
				result[i][int(not sense)] = value

			self.pool.close()
			self.pool.join()

		self.ls.remove_from_model([M], 'const')
		return [result[i] for i in range(N)]


if __name__ == '__main__':
	from cobra.io.sbml3 import read_sbml_model
	from cobamp.wrappers import COBRAModelObjectReader
	import time

	model =  read_sbml_model('/home/skapur/MEOCloud/Projectos/cobamp/examples/iAF1260_resources/original_model/Ec_iAF1260_flux2.xml')
	mor = COBRAModelObjectReader(model)

	cbm_mp = mor.to_cobamp_cbm('CPLEX')
	cbm_fast = mor.to_cobamp_cbm('CPLEX')

	init_sol = cbm_mp.optimize({1004:1}, False)
	Z0 = (1-1e-6) * init_sol.objective_value()
	cbm_mp.set_reaction_bounds(1004, lb=Z0)

	c1_time = time.time()

	pp = _ProcessPool(cpu_count())

	limits_mp = list(pp.map(cbm_mp.flux_limits, range(len(cbm_mp.reaction_names))))
	pp.close()
	pp.join()
	c2_time = time.time()
	print('Multi-threaded:',c2_time-c1_time,'seconds')

	fva = FluxVariabilityAnalysis(cbm_fast.model)
	limits_fast = fva.run(1004, False)
	c3_time = time.time()
	print('Multi-threaded fast FVA:',c3_time-c2_time,'seconds')

	error = 1e-6
	error_rx = []
	for i, lsts in enumerate(zip(limits_mp, limits_fast)):
		mpr, fr = lsts
		ld, ud = [mpr[i] - fr[i] for i in range(len(mpr))]
		if (abs(ld) > error) | (abs(ud) > error):
			error_rx.append([i, mpr, fr])

	print('Valid:',len(error_rx) == 0)


from itertools import chain
from numpy import zeros, where, array

def logical_or(x):
	return sum(x) >= 1

def logical_and(x):
	return sum(x) == len(x)


def aux_apply(fx, it):
	print(fx,it)
	args = [k for k in it if k is not None]
	return fx(args) if args else None

def convert_gpr_to_list(gpr, apply_fx=str, or_char = 'or', and_char = 'and'):
	proteins = list(
		map(
			lambda x: x.strip().replace('(', '').replace(')', ''),
			apply_fx(gpr).split(or_char)
		)
	)

	rules = [[s.strip() for s in x.split(and_char) if s.strip() != ''] for x in proteins]
	return rules


class GPREvaluator(object):
	def __init__(self, gpr_list, or_fx=logical_or, and_fx=logical_and, apply_fx=str, or_char='or', and_char='and'):
		self.__gprs = gpr_list
		self.or_fx, self.and_fx = or_fx, and_fx
		self.__or_char, self.__and_char = or_char, and_char
		self.apply_fx = apply_fx
		self.__gpr_list = [convert_gpr_to_list(g, apply_fx=self.apply_fx, or_char=or_char, and_char=and_char) for g in self.__gprs]
		self.__genes = 	tuple(set(list(chain(*chain(*self.__gpr_list)))))
		self.__gene_to_index_mapping = dict(zip(self.__genes,range(len(self.__genes))))

	def get_num_of_gprs(self):
		return len(self.__gpr_list)

	def gpr_has_string(self, index, string):
		return string in self.__gprs[index]

	def get_gpr_as_lists(self, index):
		return self.__gpr_list[index]

	def or_char(self):
		return self.__or_char

	def and_char(self):
		return self.__and_char

	def get_genes(self):
		return self.__genes

	def eval_gpr(self, index, state):
		print(self.__gpr_list[index])
		return aux_apply(self.or_fx,
				[aux_apply(self.and_fx, [state[x] for x in gs if x in state.keys()]) for gs in self.__gpr_list[index]])

	def associated_genes(self, index):
		return list(chain(*self.__gpr_list[index]))

	def associated_gene_matrix(self):
		B = zeros([len(self.__genes), len(self.__gpr_list)])
		row_ind, col_ind = [], []
		for i in range(self.get_num_of_gprs()):
			gene_indexes = [self.__gene_to_index_mapping[k] for k in self.associated_genes(i)]
			row_ind += gene_indexes
			col_ind += [i]*len(gene_indexes)
		B[row_ind, col_ind] = 1
		return B

class GeneMatrixBuilder(object):
	def __init__(self, gpr_evaluator):
		self.__gpr_evaluator = gpr_evaluator
		self.__B, self.__b = self.__get_association_matrix()

	def __get_association_matrix(self):
		B = self.__gpr_evaluator.associated_gene_matrix()
		b = B.sum(axis=0)
		return B, b

	def __step_one(self):

		# build step 1 matrices
		# G is for reactions
		# F is for genes

		single_gene_rx = where(self.__b == 1)[0]
		matched_genes = where(self.__B[:, single_gene_rx])[0]

		F1, G1 = [zeros([len(single_gene_rx), n]) for n in self.__B.shape]
		for mat, ind in zip([F1,G1],[matched_genes, single_gene_rx]):
			mat[list(range(len(single_gene_rx))), ind] = 1
		return F1, G1, single_gene_rx

	def __step_exclusive_chars(self, and_ops=True):

		# build step 2 matrices

		exclusive_char_gene_rx = array([i for i in set(where(self.__b > 1)[0]) if self.__gpr_evaluator.gpr_has_string(i, self.__gpr_evaluator.or_char() if and_ops else self.__gpr_evaluator.and_char())])
		matched_genes, rx_ind = where(self.__B[:, exclusive_char_gene_rx])
		mat_inds = [exclusive_char_gene_rx[i] for i in rx_ind]

		F1, G1 = [zeros([len(exclusive_char_gene_rx), n]) for n in self.__B.shape]
		for mat, ind in zip([F1,G1],[mat_inds, exclusive_char_gene_rx]):
			mat[list(range(len(exclusive_char_gene_rx))), ind] = 1

		return F1, G1, exclusive_char_gene_rx

	def build_gene_reaction_network(self, reactions):
		gpr_lists = [self.__gpr_evaluator.get_gpr_as_lists(i) for i in reactions]


# def __identify_single_gene_reactions(self):
	#
	# 	for i in self.__gpr_evaluator.get_num_of_gprs()
	#

if __name__ == '__main__':
	gprs = [
		'G1 and G2',
		'G3 or G4',
		'(G1 and G2) or (G3 and G4)',
		'G5',
		'G6'
	]
	#gprs = [model.reactions[i].gene_reaction_rule for i in range(len(model.reactions))]
	gpr_eval = GPREvaluator(gprs, or_fx=logical_or, and_fx=logical_and)
	for i in range(len(gprs)):
		print(gprs[i], gpr_eval.get_gpr_as_lists(i))


from .external_wrappers import CobampModelObjectReader, FramedModelObjectReader, COBRAModelObjectReader, MatFormatReader
from .method_wrappers import KShortestEFMEnumeratorWrapper, KShortestEFPEnumeratorWrapper, KShortestMCSEnumeratorWrapper
from . import external_wrappers, method_wrappers
